@@ -1,7 +1,7 @@
 /******************************************************************************
 
 	hardhat - read and write databases optimized for filename-like keys
-	Copyright (c) 2011,2012 Wessel Dankers <wsl@fruit.je>
+	Copyright (c) 2011,2012,2014-2016 Wessel Dankers <wsl@fruit.je>
 
 	This program is free software; you can redistribute it and/or modify
 	it under the terms of the GNU General Public License as published by
@@ -18,8 +18,6 @@
 
 ******************************************************************************/
 
-#include "config.h"
-
 #include <stdio.h>
 #include <stdbool.h>
 #include <stdarg.h>
@@ -32,30 +30,36 @@
 #include <errno.h>
 #include <time.h>
 #include <sys/mman.h>
-#if defined(HAVE___FPURGE) && defined(HAVE_STDIO_EXT_H)
-#include <stdio_ext.h>
-#endif
 
 #include "maker.h"
 #include "hashtable.h"
 #include "layout.h"
 
+#ifndef O_LARGEFILE
+#define O_LARGEFILE 0
+#endif
+
 /******************************************************************************
 
-	Module to create a hardhat table. Uses an unholy combination of
-	buffered I/O and low-level mmap() to get the job done. Caveat developer.
+	Module to create a hardhat database.
 
 ******************************************************************************/
 
 #define export __attribute__((visibility("default")))
 
+#define OUTBUFSIZE ((size_t)65536)
+
 struct hardhat_maker {
 	/* Output file handle */
-	FILE *db;
+	int fd;
 	/* Database file name */
 	char *filename;
 	/* Buffer used to manipulate key values (normalization, etc) */
 	uint8_t *keybuf;
+	/* Output buffer */
+	uint8_t *outbuf;
+	/* Output buffer usage */
+	size_t outbuflen;
 	/* Window into the already written data, used to detect duplicates */
 	uint8_t *window;
 	/* Size of window */
@@ -70,13 +74,15 @@ struct hardhat_maker {
 	uint32_t recnum;
 	/* Hashtable of added records, used to detect duplicates */
 	struct hashtable *hashtable;
+	/* Indicates what went wrong in case of failure */
+	char *error;
 	/* If this boolean is set, database creation has failed and
 		cannot be restarted or continued. */
 	bool failed;
+	/* Entries have been added, so the alignment is fixed now */
+	bool started;
 	/* Database is completed and cannot be modified anymore */
 	bool finished;
-	/* Indicates what went wrong in case of failure */
-	char *error;
 	/* The superblock, as it will be created at the end */
 	struct hardhat superblock;
 };
@@ -85,15 +91,16 @@ struct hardhat_maker {
 	error dynamically */
 static char enomem[] = "Out of memory";
 
+#define HARDHAT_DEFAULT_ALIGNMENT (3)
+#define HARDHAT_DEFAULT_BLOCKSIZE (12)
+
 /* struct defaults */
 static const hardhat_maker_t hardhat_maker_0 = {
+	.fd = -1,
 	.recbufsize = 65536,
-	.window = MAP_FAILED
+	.window = MAP_FAILED,
 };
 
-/* source buffer for padding */
-static const char padding[4096];
-
 /* Return the error (if any) or an empty string (but never NULL) */
 export const char *hardhat_maker_error(hardhat_maker_t *hhm) {
 	return hhm
@@ -219,24 +226,6 @@ export int hardhat_cmp(const void *a, size_t al, const void *b, size_t bl) {
 	return ac < bc ? -1 : 1;
 }
 
-/* Calculate padding for 4-byte alignment */
-static size_t pad4(size_t x) {
-	size_t p;
-	p = x & 3;
-	if(p)
-		return x + 4 - p;
-	return x;
-}
-
-/* Calculate padding for 4096-byte alignment */
-static size_t pad4k(size_t x) {
-	size_t p;
-	p = x & 4095;
-	if(p)
-		return x + 4096 - p;
-	return x;
-}
-
 /* Convenience macros to fetch aligned n-bit values */
 #define u16read(buf) (*(uint16_t *)(buf))
 #define u32read(buf) (*(uint32_t *)(buf))
@@ -244,12 +233,14 @@ static size_t pad4k(size_t x) {
 
 /* Allocate and set an error message, using printf semantics */
 static void hhm_set_error(hardhat_maker_t *hhm, const char *fmt, ...) {
-	int r;
+	int r, err;
 	va_list ap;
 
 	if(!hhm)
 		return;
 
+	err = errno;
+
 	if(hhm->error != enomem)
 		free(hhm->error);
 	hhm->error = malloc(4096);
@@ -263,6 +254,8 @@ static void hhm_set_error(hardhat_maker_t *hhm, const char *fmt, ...) {
 	} else {
 		hhm->error = enomem;
 	}
+
+	errno = err;
 }
 
 static uint32_t makeseed(void) {
@@ -299,9 +292,193 @@ static uint32_t makeseed(void) {
 	return calchash_murmur3((const void *)ts, sizeof *ts * clocks, getpid());
 }
 
+export uint64_t hardhat_maker_alignment(hardhat_maker_t *hhm, uint64_t alignment) {
+	uint64_t prev;
+
+	if(!hhm || hhm->failed)
+		return 0;
+
+	prev = UINT64_C(1) << hhm->superblock.alignment;
+
+	if(alignment) {
+		if(hhm->started)
+			return hhm_set_error(hhm, "can't change alignment after output has started"), 0;
+
+		if(alignment & (alignment - 1))
+			return hhm_set_error(hhm, "data alignment must be a power of 2"), 0;
+		hhm->superblock.alignment = ffs(alignment) - 1;
+	}
+
+	return prev;
+}
+
+export uint64_t hardhat_maker_blocksize(hardhat_maker_t *hhm, uint64_t blocksize) {
+	uint64_t prev;
+
+	if(!hhm || hhm->failed)
+		return 0;
+
+	prev = UINT64_C(1) << hhm->superblock.blocksize;
+
+	if(blocksize) {
+		if(hhm->started)
+			return hhm_set_error(hhm, "can't change blocksize after output has started"), 0;
+
+		if(blocksize & (blocksize - 1))
+			return hhm_set_error(hhm, "block size must be a power of 2"), 0;
+		hhm->superblock.blocksize = ffs(blocksize) - 1;
+	}
+
+	return prev;
+}
+
+/* Write bytes to the database and handle any errors */
+static bool hhm_db_write(hardhat_maker_t *hhm, const void *buf, size_t len) {
+	ssize_t r;
+
+	if(!len)
+		return true;
+
+	while(len) {
+		r = write(hhm->fd, buf, len);
+		switch(r) {
+			case -1:
+				hhm_set_error(hhm, "writing %zu bytes to %s failed: %m", len, hhm->filename);
+				hhm->failed = true;
+				return false;
+			case 0:
+				hhm_set_error(hhm, "writing %zu bytes to %s failed: short write", len, hhm->filename);
+				hhm->failed = true;
+				errno = EAGAIN;
+				return false;
+			default:
+				len -= r;
+				buf = (const uint8_t *)buf + r;
+		}
+	}
+
+	return true;
+}
+
+static bool hhm_db_flush(hardhat_maker_t *hhm) {
+	size_t len = hhm->outbuflen;
+	if(!len)
+		return true;
+	hhm->outbuflen = 0;
+	return hhm_db_write(hhm, hhm->outbuf, len);
+}
+
+/* Append bytes to the database and update off */
+static bool hhm_db_append(hardhat_maker_t *hhm, const void *buf, size_t len) {
+	size_t remaining;
+
+	if(!len)
+		return true;
+
+	if(len >= OUTBUFSIZE) {
+		if(!hhm_db_flush(hhm))
+			return false;
+		if(!hhm_db_write(hhm, buf, len))
+			return false;
+	} else {
+		remaining = OUTBUFSIZE - hhm->outbuflen;
+		if(len > remaining) {
+			memcpy(hhm->outbuf + hhm->outbuflen, buf, remaining);
+			if(!hhm_db_write(hhm, hhm->outbuf, OUTBUFSIZE))
+				return false;
+
+			hhm->outbuflen = len - remaining;
+			memcpy(hhm->outbuf, (const uint8_t *)buf + remaining, hhm->outbuflen);
+		} else {
+			memcpy(hhm->outbuf + hhm->outbuflen, buf, len);
+			hhm->outbuflen += len;
+			if(len == remaining && !hhm_db_flush(hhm))
+				return false;
+		}
+	}
+
+	hhm->off += len;
+	return true;
+}
+
+static bool hhm_db_seek(hardhat_maker_t *hhm, off_t off, int whence) {
+	if(!hhm_db_flush(hhm))
+		return false;
+	if(lseek(hhm->fd, off, whence) == -1) {
+		hhm_set_error(hhm, "seeking %lld bytes into %s failed: %m", (long long)off, hhm->filename);
+		hhm->failed = true;
+		return false;
+	}
+	return true;
+}
+
+static bool hhm_db_pad(hardhat_maker_t *hhm, size_t length, size_t alignment) {
+	size_t blocksize, offset, align, start, end, remaining;
+
+	blocksize = 1 << hhm->superblock.blocksize;
+	offset = hhm->off;
+
+	align = -offset % alignment;
+	offset += align;
+
+	start = offset % blocksize;
+	end = blocksize - -(offset + length) % blocksize;
+
+	if(start > end)
+		align += -offset % blocksize;
+
+	if(!align)
+		return true;
+
+	if(align >= OUTBUFSIZE) {
+		if(!hhm_db_seek(hhm, align, SEEK_CUR))
+			return false;
+	} else {
+		remaining = OUTBUFSIZE - hhm->outbuflen;
+		if(align > remaining) {
+			memset(hhm->outbuf + hhm->outbuflen, 0, remaining);
+			if(!hhm_db_write(hhm, hhm->outbuf, OUTBUFSIZE))
+				return false;
+			hhm->outbuflen = align - remaining;
+			memset(hhm->outbuf, 0, hhm->outbuflen);
+		} else {
+			memset(hhm->outbuf + hhm->outbuflen, 0, align);
+			hhm->outbuflen += align;
+			if(align == remaining && !hhm_db_flush(hhm))
+				return false;
+		}
+	}
+
+	hhm->off += align;
+	return true;
+}
+
+/* Fetch already written bytes from the database by maintaining a mmap()ed
+	window on it */
+static const uint8_t *hhm_getrec(hardhat_maker_t *hhm, uint64_t off) {
+	if(hhm->windowsize <= off) {
+		if(hhm->window != MAP_FAILED)
+			munmap(hhm->window, hhm->windowsize);
+		if(!hhm_db_flush(hhm))
+			return NULL;
+		hhm->window = mmap(NULL, hhm->off, PROT_READ, MAP_SHARED, hhm->fd, 0);
+		if(hhm->window == MAP_FAILED) {
+			hhm_set_error(hhm, "mmap()ing %s failed: %m", hhm->filename);
+			hhm->failed = true;
+			return NULL;
+		}
+		hhm->windowsize = hhm->off;
+	}
+	return hhm->window + off;
+}
+
 /* Allocate and initialize a hardhat_maker_t structure.
 	Returns NULL on failure, with errno set to the problem. */
 export hardhat_maker_t *hardhat_maker_new(const char *filename) {
+	return hardhat_maker_newat(AT_FDCWD, filename, 0666);
+}
+
+export hardhat_maker_t *hardhat_maker_newat(int dirfd, const char *filename, int mode) {
 	hardhat_maker_t *hhm;
 	int err;
 
@@ -317,6 +494,8 @@ export hardhat_maker_t *hardhat_maker_new(const char *filename) {
 	*hhm = hardhat_maker_0;
 
 	hhm->superblock.hashseed = makeseed();
+	hhm->superblock.alignment = HARDHAT_DEFAULT_ALIGNMENT;
+	hhm->superblock.blocksize = HARDHAT_DEFAULT_BLOCKSIZE;
 
 	hhm->filename = strdup(filename);
 	if(!hhm->filename) {
@@ -334,16 +513,23 @@ export hardhat_maker_t *hardhat_maker_new(const char *filename) {
 		return NULL;
 	}
 
-	hhm->db = fopen(filename, "w+");
-	if(!hhm->db) {
+	hhm->outbuf = malloc(OUTBUFSIZE);
+	if(!hhm->outbuf) {
 		err = errno;
 		hardhat_maker_free(hhm);
 		errno = err;
 		return NULL;
 	}
 
-	hhm->off = 4096;
-	if(fseek(hhm->db, hhm->off, SEEK_SET) == -1) {
+	hhm->fd = openat(dirfd, filename, O_RDWR|O_CREAT|O_LARGEFILE|O_NOCTTY|O_CLOEXEC, mode);
+	if(hhm->fd == -1) {
+		err = errno;
+		hardhat_maker_free(hhm);
+		errno = err;
+		return NULL;
+	}
+
+	if(!hhm_db_append(hhm, &hhm->superblock, sizeof hhm->superblock) || !hhm_db_flush(hhm)) {
 		err = errno;
 		hardhat_maker_free(hhm);
 		errno = err;
@@ -370,57 +556,11 @@ export hardhat_maker_t *hardhat_maker_new(const char *filename) {
 	return hhm;
 }
 
-/* Write bytes to the database and handle any errors */
-static bool hhm_db_write(hardhat_maker_t *hhm, const void *buf, size_t len) {
-	if(!len)
-		return true;
-
-	if(fwrite(buf, 1, len, hhm->db) < len) {
-		hhm_set_error(hhm, "writing %d bytes to %s failed: %m", len, hhm->filename);
-		hhm->failed = true;
-		return false;
-	}
-
-	return true;
-}
-
-/* Append bytes to the database and update off */
-static bool hhm_db_append(hardhat_maker_t *hhm, const void *buf, size_t len) {
-	if(!hhm_db_write(hhm, buf, len))
-		return false;
-	hhm->off += len;
-	return true;
-}
-
-/* Fetch already written bytes from the database by maintaining a mmap()ed
-	window on it */
-static const uint8_t *hhm_getrec(hardhat_maker_t *hhm, uint64_t off) {
-	if(hhm->windowsize <= off) {
-		if(hhm->window != MAP_FAILED)
-			munmap(hhm->window, hhm->windowsize);
-		if(fflush(hhm->db) == EOF) {
-			hhm_set_error(hhm, "writing to %s failed: %m", hhm->filename);
-			hhm->window = MAP_FAILED;
-			hhm->failed = true;
-			return NULL;
-		}
-		hhm->window = mmap(NULL, hhm->off, PROT_READ, MAP_SHARED, fileno(hhm->db), 0);
-		if(hhm->window == MAP_FAILED) {
-			hhm_set_error(hhm, "mmap()ing %s failed: %m", hhm->filename);
-			hhm->failed = true;
-			return NULL;
-		}
-		hhm->windowsize = hhm->off;
-	}
-	return hhm->window + off;
-}
 
 /* Add an entry to the database (if it doesn't exist yet).
 	Returns true on success (or if the entry already existed!)
 	Returns false on error. */
 export bool hardhat_maker_add(hardhat_maker_t *hhm, const void *key, uint16_t keylen, const void *data, uint32_t datalen) {
-	size_t recsize, padsize;
-	static const char padding[4] = {0};
 	uint32_t hash, hp, value;
 	uint64_t off;
 	struct hashtable *ht;
@@ -477,12 +617,18 @@ export bool hardhat_maker_add(hardhat_maker_t *hhm, const void *key, uint16_t ke
 			hp = 0;
 	}
 
-	/* Write out the entry to disk */
-	recsize = (size_t)6 + (size_t)keylen + (size_t)datalen;
-	padsize = pad4(recsize);
+	hhm->started = true;
+
+	/* For padding purposes, only use the size fields in the calculation. */
+	/* Using more would cause the file to increase in size significantly. */
+
+	if(!hhm_db_pad(hhm, (size_t)6 + (size_t)datalen, 4))
+		return false;
 
 	off = hhm->off;
 
+	/* Write out the entry to disk */
+
 	if(!hhm_db_append(hhm, &datalen, sizeof datalen))
 		return false;
 
@@ -492,10 +638,10 @@ export bool hardhat_maker_add(hardhat_maker_t *hhm, const void *key, uint16_t ke
 	if(!hhm_db_append(hhm, key, keylen))
 		return false;
 
-	if(!hhm_db_append(hhm, data, datalen))
+	if(!hhm_db_pad(hhm, datalen, (size_t)1 << hhm->superblock.alignment))
 		return false;
 
-	if(!hhm_db_append(hhm, padding, padsize - recsize))
+	if(!hhm_db_append(hhm, data, datalen))
 		return false;
 
 	/* Add the entry offset to the list (resizing it as necessary) */
@@ -546,59 +692,94 @@ export bool hardhat_maker_parents(hardhat_maker_t *hhm, const void *data, uint32
 	return true;
 }
 
-static hardhat_maker_t *qsort_data;
-
 /* Compare two entries from the hashtable by fetching their key values
 	and comparing them using hardhat_cmp().
 	Empty hash values always come last. */
-static int qsort_directory_cmp(const void *a, const void *b) {
+static int qsort_directory_cmp(const void *a, const void *b, void *hhm) {
 	const uint8_t *ar, *br;
 	uint32_t ad, bd;
 	const uint64_t *recs;
 
-	if(!qsort_data)
+	if(((hardhat_maker_t *)hhm)->failed)
 		return 0;
 
 	ad = ((const struct hashentry *)a)->data;
 	bd = ((const struct hashentry *)b)->data;
 
-	if(ad == UINT32_MAX)
-		return bd == UINT32_MAX ? 0 : 1;
-	else if(bd == UINT32_MAX)
+	if(ad == EMPTYHASH)
+		return bd == EMPTYHASH ? 0 : 1;
+	else if(bd == EMPTYHASH)
 		return -1;
 
-	recs = qsort_data->recbuf;
+	recs = ((hardhat_maker_t *)hhm)->recbuf;
 
-	ar = hhm_getrec(qsort_data, recs[ad]);
-	if(!ar) {
-		qsort_data = NULL;
+	ar = hhm_getrec(hhm, recs[ad]);
+	if(!ar)
 		return 0;
-	}
 
-	br = hhm_getrec(qsort_data, recs[bd]);
-	if(!br) {
-		qsort_data = NULL;
+	br = hhm_getrec(hhm, recs[bd]);
+	if(!br)
 		return 0;
-	}
 
-	/* get the first record again: it may have moved */
-	ar = hhm_getrec(qsort_data, recs[ad]);
-	if(!ar) {
-		qsort_data = NULL;
+	/* get the first record again: the memory mapping may have moved */
+	ar = hhm_getrec(hhm, recs[ad]);
+	if(!ar)
 		return 0;
-	}
 
 	return hardhat_cmp(ar + 6, u16read(ar + 4), br + 6, u16read(br + 4));
 }
 
-/* Compare hash entries by hash value */
-static int qsort_hash_cmp(const void *a, const void *b) {
-	uint32_t am, bm;
+/* Compare hash entries by hash value, with string comparison as a tie breaker. */
+static int qsort_hash_cmp(const void *a, const void *b, void *hhm) {
+	const uint8_t *ar, *br;
+	uint32_t ad, bd;
+	uint16_t al, bl;
+	const uint64_t *recs;
+	int r;
 
-	am = ((const struct hashentry *)a)->hash;
-	bm = ((const struct hashentry *)b)->hash;
+	ad = ((const struct hashentry *)a)->hash;
+	bd = ((const struct hashentry *)b)->hash;
 
-	return am < bm ? -1 : am != bm;
+	if(ad == bd) {
+		if(((hardhat_maker_t *)hhm)->failed)
+			return 0;
+
+		ad = ((const struct hashentry *)a)->data;
+		bd = ((const struct hashentry *)b)->data;
+
+		if(ad == EMPTYHASH)
+			return bd != EMPTYHASH;
+		else if(bd == EMPTYHASH)
+			return -1;
+
+		recs = ((hardhat_maker_t *)hhm)->recbuf;
+
+		ar = hhm_getrec(hhm, recs[ad]);
+		if(!ar)
+			return 0;
+
+		br = hhm_getrec(hhm, recs[bd]);
+		if(!br)
+			return 0;
+
+		/* get the first record again: the memory mapping may have moved */
+		ar = hhm_getrec(hhm, recs[ad]);
+		if(!ar)
+			return 0;
+
+		al = u16read(ar + 4);
+		bl = u16read(br + 4);
+
+		if(al < bl) {
+			r = memcmp(ar + 6, br + 6, al);
+			return r ? r : -1;
+		} else {
+			r = memcmp(ar + 6, br + 6, bl);
+			return r ? r : al != bl;
+		}
+	}
+
+	return ad < bd ? -1 : 1;
 }
 
 /* Find the longest common prefix (on ‘/’ boundaries) */
@@ -621,9 +802,13 @@ static size_t common_parents(const uint8_t *a, size_t al, const uint8_t *b, size
 	return cl;
 }
 
+#ifndef HAVE_QSORT_R
+extern void qsort_r(void *, size_t, size_t, int (*)(const void *, const void *, void *), void *);
+#endif
+
 /* Finish up the database by writing the indexes and the superblock */
 export bool hardhat_maker_finish(hardhat_maker_t *hhm) {
-	FILE *db;
+	int fd;
 	struct hashtable *ht;
 	struct hashentry *he;
 	uint32_t i, num, pfxnum;
@@ -636,24 +821,17 @@ export bool hardhat_maker_finish(hardhat_maker_t *hhm) {
 		return false;
 	}
 
-	db = hhm->db;
 	num = hhm->recnum;
-
 	hhm->superblock.data_end = hhm->off;
 
-	if(!hhm_db_append(hhm, padding, pad4k(hhm->off) - hhm->off)) {
-		hhm->failed = true;
-		return false;
-	}
-
 	/* Sort the hashtable in directory order */
 	ht = hhm->hashtable;
-	qsort_data = hhm;
-	qsort(ht->buf, ht->size, sizeof *ht->buf, qsort_directory_cmp);
-	if(!qsort_data) {
-		hhm->failed = true;
+	qsort_r(ht->buf, ht->size, sizeof *ht->buf, qsort_directory_cmp, hhm);
+	if(hhm->failed)
+		return false;
+
+	if(!hhm_db_pad(hhm, num * sizeof *dir, sizeof *dir))
 		return false;
-	}
 
 	hhm->superblock.directory_start = hhm->off;
 
@@ -670,13 +848,19 @@ export bool hardhat_maker_finish(hardhat_maker_t *hhm) {
 
 	hhm->superblock.directory_end = hhm->off;
 
+	if(!hhm_getrec(hhm, hhm->superblock.directory_end))
+		return false;
+
+	/* Read back the list of offsets as we wrote it out earlier */
+	memcpy(dir, hhm->window + hhm->superblock.directory_start, sizeof *dir * num);
+
 	/* Now sort the hashtable again, this time on hash value */
-	qsort(ht->buf, num, sizeof *ht->buf, qsort_hash_cmp);
+	qsort_r(ht->buf, num, sizeof *ht->buf, qsort_hash_cmp, hhm);
+	if(hhm->failed)
+		return false;
 
-	if(!hhm_db_append(hhm, padding, pad4k(hhm->off) - hhm->off)) {
-		hhm->failed = true;
+	if(!hhm_db_pad(hhm, num * sizeof *ht->buf, sizeof *ht->buf))
 		return false;
-	}
 
 	hhm->superblock.hash_start = hhm->off;
 
@@ -687,17 +871,6 @@ export bool hardhat_maker_finish(hardhat_maker_t *hhm) {
 
 	hhm->superblock.hash_end = hhm->off;
 
-	if(!hhm_db_append(hhm, padding, pad4k(hhm->off) - hhm->off)) {
-		hhm->failed = true;
-		return false;
-	}
-
-	if(!hhm_getrec(hhm, hhm->superblock.directory_end))
-		return false;
-
-	/* Read back the list of offsets as we wrote it out earlier */
-	memcpy(dir, hhm->window + hhm->superblock.directory_start, sizeof *dir * num);
-
 	/* Calculate the list of common prefixes, reusing the old hash
 		table as storage */
 	prev = NULL;
@@ -739,7 +912,12 @@ export bool hardhat_maker_finish(hardhat_maker_t *hhm) {
 	}
 
 	/* Write out the prefix list as a hash table */
-	qsort(ht->buf, pfxnum, sizeof *ht->buf, qsort_hash_cmp);
+	qsort_r(ht->buf, pfxnum, sizeof *ht->buf, qsort_hash_cmp, hhm);
+	if(hhm->failed)
+		return false;
+
+	if(!hhm_db_pad(hhm, pfxnum  * sizeof *ht->buf, sizeof *ht->buf))
+		return false;
 
 	hhm->superblock.prefix_start = hhm->off;
 
@@ -748,51 +926,40 @@ export bool hardhat_maker_finish(hardhat_maker_t *hhm) {
 
 	hhm->superblock.prefix_end = hhm->off;
 
-	if(!hhm_db_append(hhm, padding, pad4k(hhm->off) - hhm->off)) {
-		hhm->failed = true;
-		return false;
-	}
-
 	/* Create and write out the superblock */
 	memcpy(hhm->superblock.magic, HARDHAT_MAGIC, sizeof hhm->superblock.magic);
 	hhm->superblock.byteorder = UINT64_C(0x0123456789ABCDEF);
-	hhm->superblock.version = UINT32_C(2);
+	hhm->superblock.version = UINT32_C(3);
 	hhm->superblock.entries = num;
 	hhm->superblock.prefixes = pfxnum;
 	hhm->superblock.filesize = hhm->off;
 	hhm->superblock.checksum = calchash_murmur3((const void *)&hhm->superblock, sizeof hhm->superblock - 4, hhm->superblock.hashseed);
 
-	if(fseek(db, 0, SEEK_SET) == -1) {
-		hhm_set_error(hhm, "seeking in %s failed: %m", hhm->filename);
-		hhm->failed = true;
+	if(!hhm_db_seek(hhm, 0, SEEK_SET))
 		return false;
-	}
 
 	if(!hhm_db_write(hhm, &hhm->superblock, sizeof hhm->superblock))
 		return false;
 
-	hhm->db = NULL;
-	if(fflush(db) == EOF) {
-		hhm_set_error(hhm, "writing to %s failed: %m", hhm->filename);
+	if(!hhm_db_flush(hhm))
+		return false;
+
+	fd = hhm->fd;
+	if(ftruncate(fd, (off_t)hhm->off) == -1) {
+		hhm_set_error(hhm, "truncating %s failed: %m", hhm->filename);
 		hhm->failed = true;
 		return false;
 	}
 
-	if(ftruncate(fileno(db), (off_t)hhm->off) == -1) {
-		hhm_set_error(hhm, "padding %s failed: %m", hhm->filename);
-		hhm->failed = true;
-		return false;
-	}
-
-	if(fdatasync(fileno(db)) == -1) {
+	if(fdatasync(fd) == -1) {
 		hhm_set_error(hhm, "writing %s failed: %m", hhm->filename);
 		hhm->failed = true;
 		return false;
 	}
 
-	hhm->db = NULL;
-	if(fclose(db) == EOF) {
-		hhm_set_error(hhm, "writing to %s failed: %m", hhm->filename);
+	hhm->fd = -1;
+	if(close(fd) == -1) {
+		hhm_set_error(hhm, "closing %s failed: %m", hhm->filename);
 		hhm->failed = true;
 		return false;
 	}
@@ -806,26 +973,12 @@ export bool hardhat_maker_finish(hardhat_maker_t *hhm) {
 export void hardhat_maker_free(hardhat_maker_t *hhm) {
 	if(!hhm)
 		return;
-	if(hhm->db) {
-		/* Various gambits to prevent flushing the buffer.
-		** Unfinished shutdown may be due to a fork(), in which case
-		** our flushing could theoretically corrupt the result. */
-#if defined(HAVE_FPURGE)
-		fpurge(hhm->db);
-#elif defined(HAVE___FPURGE)
-		__fpurge(hhm->db);
-#elif defined(HAVE_FILENO)
-		int fd = open("/dev/null", O_NOCTTY|O_WRONLY);
-		if(fd != -1) {
-			dup2(fd, fileno(hhm->db));
-			close(fd);
-		}
-#endif
-		fclose(hhm->db);
-	}
+	if(hhm->fd != -1)
+		close(hhm->fd);
 	freehash(hhm->hashtable);
 	free(hhm->keybuf);
 	free(hhm->recbuf);
+	free(hhm->outbuf);
 	free(hhm->filename);
 	if(hhm->window != MAP_FAILED)
 		munmap(hhm->window, hhm->windowsize);
